'use strict';

var assert = require('assert');
var Devebot = require('devebot');
var Promise = Devebot.require('bluebird');
var lodash = Devebot.require('lodash');
var pinbug = Devebot.require('pinbug');
var debugx = pinbug('devebot-co-rabbitmq:handler');
var debug0 = pinbug('trace:devebot-co-rabbitmq:handler');
var debug2 = pinbug('devebot-co-rabbitmq:handler:extract');
var amqp = require('amqplib/callback_api');
var valvekit = require('valvekit');
var util = require('util');
var Readable = require('stream').Readable;
var Writable = require('stream').Writable;
var zapper = require('./zapper');

function Handler(a) {
  var b = this.logger || zapper.getLogger();
  var c = this.tracer || zapper.getTracer();

  var d = {};
  d.uri = a.uri || a.host;
  d.exchangeName = a.exchangeName || a.exchange;
  d.exchangeType = a.exchangeType || 'fanout';
  if (zapper.isNumber(a.exchangeQuota)) {
    d.exchangeQuota = a.exchangeQuota;
  }
  if (zapper.isBoolean(a.exchangeMutex)) {
    d.exchangeMutex = a.exchangeMutex;
  }
  d.durable = zapper.isBoolean(a.durable) ? a.durable : true;
  d.autoDelete = zapper.isBoolean(a.autoDelete) ? a.autoDelete : false;
  if (zapper.isString(a.alternateExchange)) {
    d.alternateExchange = a.alternateExchange;
  }
  d.routingKey = zapper.isString(a.routingKey) ? a.routingKey : 'default-routing-key';

  // "inbox" queue configuration
  var e = a.inbox || a.consumer;
  if (lodash.isObject(e)) {
    d.inbox = { uri: d.uri };
    lodash.merge(d.inbox, e);
  }

  if (!zapper.isString(d.inbox.queueName)) {
    d.inbox.queueName = a.queueName || a.queue || '';
  }
  if (!zapper.isBoolean(d.inbox.durable)) {
    d.inbox.durable = zapper.isBoolean(a.durable) ? a.durable : true;
  }
  if (!zapper.isBoolean(d.inbox.exclusive)) {
    d.inbox.exclusive = zapper.isBoolean(a.exclusive) ? a.exclusive : false;
  }
  if (!zapper.isBoolean(d.inbox.noAck)) {
    d.inbox.noAck = zapper.isBoolean(a.noAck) ? a.noAck : false;
  }
  if (!zapper.isNumber(d.inbox.prefetch)) {
    d.inbox.prefetch = zapper.isNumber(a.prefetch) ? a.prefetch : undefined;
  }
  if (!zapper.isBoolean(d.inbox.connectIsCached)) {
    d.inbox.connectIsCached = zapper.isBoolean(a.connectIsCached) ? a.connectIsCached : true;
  }
  if (!zapper.isBoolean(d.inbox.channelIsCached)) {
    d.inbox.channelIsCached = zapper.isBoolean(a.channelIsCached) ? a.channelIsCached : true;
  }
  if (!zapper.isBoolean(d.inbox.noBinding)) {
    d.inbox.noBinding = zapper.isBoolean(a.noBinding) ? a.noBinding : false;;
  }
  if (!zapper.isNumber(d.inbox.maxSubscribers)) {
    d.inbox.maxSubscribers = a.maxSubscribers;
  }

  // "trash" queue configuration
  var f = a.trash || a.recycler;
  if (lodash.isObject(f)) {
    d.trash = {
      uri: d.uri,
      redeliveredCountName: 'x-redelivered-count',
      redeliveredLimit: 5
    };
    lodash.merge(d.trash, f);
    if (lodash.isEmpty(d.trash.queueName) && d.inbox && d.inbox.queueName) {
      d.trash.queueName = d.inbox.queueName + '-trashed';
    }
  }

  d.maxListeners = 20;
  if (zapper.isPositiveInteger(a.maxListeners)) {
    d.maxListeners = a.maxListeners;
  }
  if (d.inbox) {
    d.inbox.maxListeners = d.inbox.maxListeners || d.maxListeners;
  }
  if (d.chain) {
    d.chain.maxListeners = d.chain.maxListeners || d.maxListeners;
  }
  if (d.trash) {
    d.trash.maxListeners = d.trash.maxListeners || d.maxListeners;
  }

  debugx.enabled && debugx(' - configuration object: %s', JSON.stringify(d));

  Object.defineProperty(this, 'config', {
    get: function a() {
      return d;
    },
    set: function b(a) {}
  });
};

var d = function d(a) {
  var b = a.state;

  var c = void 0;
  if (b.guard) {
    c = new Promise(function (a, c) {
      b.guard.lock(function () {
        a();
      });
    });
    c.finally(function () {
      b.guard.unlock();
    });
  } else {
    c = Promise.resolve();
  }
  return c;
};

var d = function d(a) {
  var b = a.state,
      c = a.config;


  b.connectionCount = b.connectionCount || 0;
  debugx.enabled && debugx('getConnection() - connection amount: %s', b.connectionCount);

  if (c.connectIsCached !== false && b.connection) {
    debugx.enabled && debugx('getConnection() - connection has been available');
    return Promise.resolve(b.connection);
  } else {
    debugx.enabled && debugx('getConnection() - make a new connection');
    var e = Promise.promisify(amqp.connect, { context: amqp });
    return e(c.uri, {}).then(function (a) {
      b.connectionCount += 1;
      if (c.maxListeners > 0) {
        a.setMaxListeners(c.maxListeners);
      }
      a.on('close', function () {
        debugx.enabled && debugx('getConnection() - connection is closed');
        delete b.connection;
        b.connectionCount--;
      });
      a.on('error', function (a) {
        debugx.enabled && debugx('getConnection() - connection has error');
        delete b.connection;
      });
      debugx.enabled && debugx('getConnection() - connection is created successfully');
      return b.connection = a;
    }).catch(function (b) {
      return Promise.resolve(a).delay(1100).then(getConnection);
    });
  }
};

var d = function d(a) {
  var b = a.state,
      c = a.config;

  if (c.channelIsCached !== false && b.channel) {
    debugx.enabled && debugx('getChannel() - channel has been available');
    return Promise.resolve(b.channel);
  }
  return getTicket(a).then(function () {
    if (c.channelIsCached !== false && b.channel) {
      debugx.enabled && debugx('getChannel() - channel has been available');
      return Promise.resolve(b.channel);
    } else {
      debugx.enabled && debugx('getChannel() - make a new channel');
      return getConnection(a).then(function (c) {
        debugx.enabled && debugx('getChannel() - connection has already');
        var d = Promise.promisify(c.createConfirmChannel, { context: c });
        return d().then(function (a) {
          a.on('close', function () {
            debugx.enabled && debugx('getChannel() - channel is closed');
            delete b.channel;
          });
          a.on('error', function (a) {
            debugx.enabled && debugx('getChannel() - channel has error');
            delete b.channel;
          });
          a.on('drain', function () {
            debugx.enabled && debugx('getChannel() - channel has drained, open the valve');
            b.valve && b.valve.set(true);
          });
          debugx.enabled && debugx('getChannel() - channel is created successfully');
          return b.channel = a;
        }).catch(function (b) {
          return Promise.resolve(a).delay(1700).then(getChannel);
        });
      });
    }
  });
};

var d = function d(a, b) {
  var c = a.config;

  b = b || {};
  return getChannel(a).then(function (d) {
    var e = assertExchange(a, d);
    return e.then(function (a) {
      return {
        exchangeName: a.exchange || c.exchangeName,
        routingKey: b.routingKey || c.routingKey
      };
    });
  });
};

var f = function f(a, b) {
  var c = a.state,
      d = a.config;

  if (c.exchangeAsserted) return Promise.resolve(c.exchangeAsserted);
  var e = Promise.promisify(b.assertExchange, { context: b });
  return e(d.exchangeName, d.exchangeType, {
    durable: d.durable,
    autoDelete: d.autoDelete
  }).then(function (a) {
    return c.exchangeAsserted = a;
  });
};

var g = function g(a, b, c) {
  var d = a.state,
      e = a.config;

  if (!c && d.queueAsserted) return Promise.resolve(d.queueAsserted);
  var f = {
    durable: e.durable,
    exclusive: e.exclusive,
    autoDelete: e.autoDelete
  };
  if (zapper.isInteger(e.messageTtl) && e.messageTtl >= 0) f.messageTtl = e.messageTtl;
  if (zapper.isPositiveInteger(e.expires)) f.expires = e.expires;
  if (zapper.isPositiveInteger(e.maxLength)) f.maxLength = e.maxLength;
  if (zapper.isPositiveInteger(e.maxPriority)) f.maxPriority = e.maxPriority;
  if (zapper.isString(e.deadLetterExchange)) f.deadLetterExchange = e.deadLetterExchange;
  if (zapper.isString(e.deadLetterRoutingKey)) f.deadLetterRoutingKey = e.deadLetterRoutingKey;
  return Promise.promisify(b.assertQueue, { context: b })(e.queueName, f).then(function (a) {
    return d.queueAsserted = a;
  });
};

var c = function c(a, b) {
  return Promise.promisify(b.checkQueue, { context: b })(a.config.queueName);
};

var c = function c(a, b) {
  return Promise.promisify(b.purgeQueue, { context: b })(a.config.queueName);
};

var f = function f(a, b, c) {
  var d = a.state,
      e = a.config;

  c = c || {};
  debugx.enabled && debugx('%s() an object to rabbitmq queue', a.name);
  return getChannel(a).then(function (f) {
    return assertQueue(a, f).then(function (g) {
      var h = d.connection;
      return new Promise(function (g, i) {
        var a = function a() {
          i({ msg: 'Timeout exception' });
        };
        h.on('error', j);
        try {
          checkValve(a, function () {
            var a = f.sendToQueue(e.queueName, zapper.bufferify(b), c, function (a, b) {
              h.removeListener('error', j);
              if (a) {
                i(a);
              } else {
                b = b || {};
                b.drained = true;
                g(b);
              }
            });
            if (!a) {
              debugx.enabled && debugx('channel is overflowed, close the valve');
              d.valve && d.valve.set(false);
            }
          });
        } catch (a) {
          h.removeListener('error', j);
          i(a);
        }
      });
    });
  }).catch(function (e) {
    debugx.enabled && debugx('sendToQueue() has failed, recall ...');
    d.connection = null;
    d.channel = null;
    return sendToQueue(a, b, c);
  });
};

var d = function d(a, b) {
  var c = a.state;

  if (c.valve) {
    c.valve.wait(function a(b) {
      return b === true;
    }, b);
  } else {
    b();
  }
};

var e = function e(a, b, c, d) {
  return getChannel(a).then(function (e) {
    return new Promise(function (f, g) {
      var a = function a() {
        g({ msg: 'Timeout exception' });
      };
      var i = a.state.connection;
      i.on('error', h);
      try {
        checkValve(a, function () {
          var j = e.publish(d.exchangeName, d.routingKey, zapper.bufferify(b), c, function (a, b) {
            i.removeListener('error', h);
            if (a) {
              g(a);
              debugx.enabled && debugx('sendToExchange() failed: %s', JSON.stringify(a));
            } else {
              b = b || {};
              b.drained = true;
              f(b);
              debugx.enabled && debugx('sendToExchange() is ok');
            }
          });
          if (!j) {
            debugx.enabled && debugx('channel is overflowed, close the valve');
            a.state.valve && a.state.valve.set(false);
          }
        });
      } catch (a) {
        i.removeListener('error', h);
        debugx.enabled && debugx('sendToExchange() throw exception: %s', JSON.stringify(a));
        g(a);
      }
    });
  }).catch(function (e) {
    debugx.enabled && debugx('produce() - sendToExchange() failed, recall ...');
    a.state.connection = null;
    a.state.channel = null;
    return sendToExchange(a, b, c, d);
  });
};

var b = function b(a) {
  a = a || this;
  a.producerState = a.producerState || {};
  a.producerState.guard = a.producerState.guard || valvekit.createMutex();
  a.producerState.valve = a.producerState.valve || valvekit.createBarrier(true);
  if (a.config.exchangeMutex) {
    a.producerState.fence = a.producerState.fence || valvekit.createBarrier(true);
  }
  if (a.config.exchangeQuota) {
    a.producerState.quota = a.producerState.quota || valvekit.createSemaphore(a.config.exchangeQuota);
  }
  return a.producerState;
};

var d = function d(a) {
  a = a || this;
  var b = getProducerState(a);
  var c = Promise.resolve(b);
  if (b.fence) {
    c = c.then(function (a) {
      return new Promise(function (b, c) {
        a.fence.wait(function a(b) {
          return b === true;
        }, function c() {
          b(a);
        });
      });
    });
  }
  if (b.quota) {
    c = c.then(function (a) {
      return new Promise(function (b, c) {
        a.quota.take(function c() {
          b(a);
        });
      });
    });
  }
  return c;
};

var c = function c(a) {
  a = a || this;
  var b = getProducerState(a);
  if (b.quota) {
    b.quota.release();
  }
};

Handler.prototype.ready = function () {
  var a = this;
  var b = { config: a.config, state: getProducerState(a) };
  return getChannel(b).then(function (c) {
    var d = assertExchange(b, c);
    if (hasInbox(a)) {
      d = d.then(function (b) {
        return resolveInbox(a).then(function (d) {
          return assertQueue(d, c).then(function (e) {
            if (d.config.noBinding === true) return e;
            return bindInbox(c, e, b, a.config.routingKey);
          });
        });
      });
    }
    if (hasTrash(a)) {
      d = d.then(function (b) {
        return resolveTrash(a).then(function (a) {
          return assertQueue(a, c);
        });
      });
    }
    if (hasChain(a)) {
      d = d.then(function (b) {
        return resolveChain(a).then(function (a) {
          return assertQueue(a, c);
        });
      });
    }
    return d;
  });
};

Handler.prototype.open = Handler.prototype.ready;
Handler.prototype.prepare = Handler.prototype.ready;

Handler.prototype.produce = function (a, b, c) {
  var d = this;
  b = b || {};
  debugx.enabled && debugx('produce() an object to rabbitmq');
  return lockProducer(d).then(function (e) {
    var f = { config: d.config, state: e };
    return retrieveExchange(f, c).then(function (c) {
      return sendToExchange(f, a, b, c);
    });
  }).finally(function () {
    unlockProducer(d);
  });
};

Handler.prototype.enqueue = Handler.prototype.produce;
Handler.prototype.publish = Handler.prototype.produce;

var b = function b(a) {
  a = a || this;
  a.pipelineState = a.pipelineState || {};
  a.pipelineState.guard = a.pipelineState.guard || valvekit.createMutex();
  a.pipelineState.valve = a.pipelineState.valve || valvekit.createBarrier(true);
  if (a.config.exchangeMutex) {
    a.pipelineState.mutex = a.pipelineState.mutex || valvekit.createMutex();
  }
  return a.pipelineState;
};

var d = function d(a) {
  a = a || this;
  var b = getProducerState(a);
  if (b.fence) {
    b.fence.set(false);
  }
  var c = getPipelineState(a);
  return new Promise(function (a, b) {
    if (c.mutex) {
      c.mutex.lock(function () {
        a(c);
      });
    } else {
      a(c);
    }
  });
};

var d = function d(a) {
  a = a || this;
  var b = getProducerState(a);
  var c = getPipelineState(a);
  b.fence && b.fence.set(true);
  c.mutex && c.mutex.unlock();
};

function ProducerStream(a, b, c) {
  b = b || {};
  Writable.call(this, { objectMode: b.objectMode !== false });

  var d = 0;
  this.count = function () {
    return d;
  };

  this._write = function (e, f, g) {
    debug2.enabled && debug2('ProducerStream._write() - sending #%s', d + 1);
    sendToExchange(a, e, b, c).then(function () {
      d++;
      debug2.enabled && debug2('ProducerStream._write() - #%s has been sent', d);
      g();
    });
  };
}

util.inherits(ProducerStream, Writable);

Handler.prototype.exhaust = function (a, b, c) {
  var d = this;
  b = b || {};
  debugx.enabled && debugx('exhaust() data from a readable stream');
  if (!(a instanceof Readable)) return Promise.reject({
    message: '[source] is not a readable stream'
  });
  return lockPipeline(d).then(function (e) {
    var f = { config: d.config, state: e };
    return retrieveExchange(f, c).then(function (c) {
      var d = new ProducerStream(f, b, c);
      return new Promise(function (b, c) {
        debug2.enabled && debug2('exhaust() - starting...');
        d.on('error', function (a) {
          debug2.enabled && debug2('exhaust() - stream error: %s', JSON.stringify(a));
          c(a);
        });
        d.on('finish', function () {
          debug2.enabled && debug2('exhaust() - stream end. Total: %s', d.count());
          b();
        });
        a.pipe(d, { end: true });
      });
    });
  }).then(function (a) {
    debug2.enabled && debug2('exhaust() - finish successfully');
    return a;
  }).catch(function (a) {
    debug2.enabled && debug2('exhaust() - error occurred: %s', JSON.stringify(a));
    return Promise.reject(a);
  }).finally(function () {
    unlockPipeline(d);
  });
};

var b = function b(a) {
  a = a || this;
  a.inboxState = a.inboxState || {};
  a.inboxState.valve = a.inboxState.valve || valvekit.createBarrier(true);
  return a.inboxState;
};

var b = function b(a) {
  a = a || this;
  return a.config && a.config.inbox && a.config.inbox.enabled !== false;
};

var b = function b(a) {
  a = a || this;
  if (!hasInbox(a)) return Promise.reject({ message: 'The inbox is unavailable' });
  return Promise.resolve({ name: 'enqueueInbox', config: a.config.inbox, state: getInboxState(a) });
};

var d = function d(a, b, c) {
  return resolveInbox(a).then(function (a) {
    return sendToQueue(a, b, c);
  });
};

var f = function f(a, b, c, d) {
  var e = Promise.promisify(a.bindQueue, { context: a });
  return e(b.queue, c.exchange, d, {}).then(function (a) {
    debugx.enabled && debugx('bindInbox() - queue: %s has been bound to %s', b.queue, c.exchange);
    return b;
  });
};

Handler.prototype.checkInbox = function () {
  return resolveInbox(this).then(function (a) {
    return getChannel(a).then(function (b) {
      return checkQueue(a, b);
    });
  });
};

Handler.prototype.purgeInbox = function () {
  return resolveInbox(this).then(function (a) {
    return getChannel(a).then(function (b) {
      return assertQueue(a, b).then(function () {
        return purgeQueue(a, b);
      });
    });
  });
};

Handler.prototype.consume = function (a) {
  var b = this;

  assert.ok(zapper.isFunction(a), 'callback should be a function');

  var c = { config: b.config.inbox, state: getInboxState(b) };
  c.state.count = c.state.count || 0;

  debugx.enabled && debugx('consume() - subscribe the Inbox queue');
  return getChannel(c).then(function (d) {

    if (c.config.prefetch && c.config.prefetch >= 0) {
      debugx.enabled && debugx('consume() - set channel prefetch: %s', c.config.prefetch);
      d.prefetch(c.config.prefetch, true);
    }

    var e = assertQueue(c, d, true);

    e = e.then(function (a) {
      debugx.enabled && debugx('consume() - queue info: %s', JSON.stringify(a));
      if (c.config.maxSubscribers && c.config.maxSubscribers <= a.consumerCount) {
        var b = {
          consumerCount: a.consumerCount,
          maxSubscribers: c.config.maxSubscribers,
          message: 'exceeding quota limits of subscribers'
        };
        debugx.enabled && debugx('consume() - Error: %s', JSON.stringify(b));
        return Promise.reject(b);
      }
      return a;
    });

    e = e.then(function (a) {
      if (c.config.noBinding === true) return a;
      return assertExchange({ config: b.config, state: c.state }, d).then(function (c) {
        var e = Promise.promisify(d.bindQueue, { context: d });
        return e(a.queue, c.exchange, b.config.routingKey, {}).then(function () {
          debugx.enabled && debugx('consume() - queue: %s has been bound', a.queue);
          return a;
        });
      });
    });

    e = e.then(function (e) {
      var f = Promise.promisify(d.consume, { context: d });
      return f(e.queue, function (e) {
        c.state.count++;
        debug0.enabled && debug0('consume() - received message: %s, fields: %s, amount: %s', e.content, JSON.stringify(e.fields), c.state.count);
        a(e.content, {
          fields: e.fields,
          properties: e.properties
        }, function a(f, g) {
          if (c.config.noAck !== true) {
            if (f) {
              var h = void 0;
              var i = e && e.content && zapper.bufferify(e.content) || null;
              var j = e && e.properties && lodash.clone(e.properties) || {};
              var k = b.config.trash;
              var l = j.headers || {};
              l[k.redeliveredCountName] = (l[k.redeliveredCountName] || 0) + 1;
              if (l[k.redeliveredCountName] <= k.redeliveredLimit) {
                debugx.enabled && debugx('consume() - enqueueInbox message');
                h = enqueueInbox(b, i, j);
              } else {
                if (hasTrash(b)) {
                  debugx.enabled && debugx('consume() - enqueueTrash message');
                  h = enqueueTrash(b, i, j);
                } else {
                  h = Promise.resolve();
                }
              }
              h.then(function () {
                d.nack(e, false, false);
              }).catch(function (a) {
                d.nack(e, false, true);
              });
            } else {
              if (hasChain(b)) {
                var m = g && g.content && zapper.bufferify(g.content);
                m = m || e && e.content && zapper.bufferify(e.content) || null;
                var n = g && g.properties && lodash.clone(g.properties);
                n = n || e && e.properties && lodash.clone(e.properties) || {};
                enqueueChain(b, m, n).then(function () {
                  d.ack(e);
                }).catch(function (a) {
                  d.nack(e, false, true);
                });
              } else {
                d.ack(e);
              }
            }
          }
          c.state.count--;
        });
      }, { noAck: c.config.noAck === true });
    });

    return e.then(function (a) {
      debugx.enabled && debugx('consume() - consumerTag: %s. CTRL+C to exit', a.consumerTag);
      c.state.consumerRefs = c.state.consumerRefs || [];
      c.state.consumerRefs.push(a);
      return a;
    });
  });
};

Handler.prototype.process = Handler.prototype.consume;

var b = function b(a) {
  a = a || this;
  a.chainState = a.chainState || {};
  a.chainState.valve = a.chainState.valve || valvekit.createBarrier(true);
  return a.chainState;
};

var b = function b(a) {
  a = a || this;
  return a.config && a.config.chain && a.config.chain.enabled !== false;
};

var b = function b(a) {
  a = a || this;
  if (!hasChain(a)) return Promise.reject({ message: 'The chain is unavailable' });
  return Promise.resolve({ name: 'enqueueChain', config: a.config.chain, state: getChainState(a) });
};

var d = function d(a, b, c) {
  return resolveChain(a).then(function (a) {
    return sendToQueue(a, b, c);
  });
};

Handler.prototype.checkChain = function () {
  return resolveChain(this).then(function (a) {
    return getChannel(a).then(function (b) {
      return checkQueue(a, b);
    });
  });
};

Handler.prototype.purgeChain = function () {
  return resolveChain(this).then(function (a) {
    return getChannel(a).then(function (b) {
      return assertQueue(a, b).then(function () {
        return purgeQueue(a, b);
      });
    });
  });
};

var b = function b(a) {
  a = a || this;
  a.trashState = a.trashState || {};
  a.trashState.valve = a.trashState.valve || valvekit.createBarrier(true);
  return a.trashState;
};

var b = function b(a) {
  a = a || this;
  return a.config && a.config.trash && a.config.trash.enabled !== false;
};

var b = function b(a) {
  a = a || this;
  if (!hasTrash(a)) return Promise.reject({ message: 'The trash is unavailable' });
  return Promise.resolve({ name: 'enqueueTrash', config: a.config.trash, state: getTrashState(a) });
};

var d = function d(a, b, c) {
  return resolveTrash(a).then(function (a) {
    return sendToQueue(a, b, c);
  });
};

Handler.prototype.checkTrash = function () {
  return resolveTrash(this).then(function (a) {
    return getChannel(a).then(function (b) {
      return checkQueue(a, b);
    });
  });
};

Handler.prototype.purgeTrash = function () {
  return resolveTrash(this).then(function (a) {
    return getChannel(a).then(function (b) {
      return assertQueue(a, b).then(function () {
        return purgeQueue(a, b);
      });
    });
  });
};

Handler.prototype.recycle = function (a) {
  var b = this;

  assert.ok(zapper.isFunction(a), 'callback should be a function');

  return resolveTrash(b).then(function (b) {
    b.state.count = b.state.count || 0;

    debugx.enabled && debugx('recycle() - get an object from Trash');
    return getChannel(b).then(function (c) {

      if (b.config.prefetch && b.config.prefetch >= 0) {
        debugx.enabled && debugx('recycle() - set channel prefetch: %s', b.config.prefetch);
        c.prefetch(b.config.prefetch, true);
      }

      var d = assertQueue(b, c, true);

      d = d.then(function (a) {
        debugx.enabled && debugx('recycle() - queue info: %s', JSON.stringify(a));
        if (b.config.maxSubscribers && b.config.maxSubscribers <= a.consumerCount) {
          var c = {
            consumerCount: a.consumerCount,
            maxSubscribers: b.config.maxSubscribers,
            message: 'exceeding quota limits of subscribers'
          };
          debugx.enabled && debugx('recycle() - Error: %s', JSON.stringify(c));
          return Promise.reject(c);
        }
        return a;
      });

      d = d.then(function (d) {
        return Promise.promisify(c.consume, { context: c })(d.queue, function (d) {
          b.state.count++;
          debugx.enabled && debugx('recycle() - received message: %s, fields: %s, amount: %s', d.content, JSON.stringify(d.fields), b.state.count);
          a(d.content, {
            fields: d.fields,
            properties: d.properties
          }, function a(e) {
            if (b.config.noAck !== true) {
              e ? c.nack(d) : c.ack(d);
            }
            b.state.count--;
          });
        }, { noAck: b.config.noAck === true });
      });

      return d.then(function (a) {
        debugx.enabled && debugx('recycle() - consumerTag: %s. CTRL+C to exit', a.consumerTag);
        b.state.consumerRefs = b.state.consumerRefs || [];
        b.state.consumerRefs.push(a);
        return a;
      });
    });
  });
};

var b = function b(a) {
  a = a || this;
  return resolveTrash(a).then(function (a) {
    if (a.state.garbage) return a.state.garbage;
    return getChannel(a).then(function (b) {
      return assertQueue(a, b).then(function (a) {
        return Promise.promisify(b.get, { context: b })(a.queue, {});
      }).then(function (b) {
        debugx.enabled && debugx('examineGarbage() - msg: %s', JSON.stringify(b));
        if (b !== false) a.state.garbage = b;
        return a.state.garbage;
      });
    });
  });
};

var garbageAction = {};

var discardGarbage = garbageAction['discard'] = function (a) {
  a = a || this;
  return resolveTrash(a).then(function (a) {
    if (!a.state.garbage) return false;
    return getChannel(a).then(function (b) {
      debugx.enabled && debugx('discardGarbage() - nack()');
      b.nack(a.state.garbage, false, false);
      a.state.garbage = undefined;
      return true;
    });
  });
};

var restoreGarbage = garbageAction['restore'] = function (a) {
  a = a || this;
  return resolveTrash(a).then(function (a) {
    if (!a.state.garbage) return false;
    return getChannel(a).then(function (b) {
      debugx.enabled && debugx('restoreGarbage() - nack()');
      b.nack(a.state.garbage);
      a.state.garbage = undefined;
      return true;
    });
  });
};

var recoverGarbage = garbageAction['recover'] = function (a) {
  a = a || this;
  return resolveTrash(a).then(function (b) {
    if (!b.state.garbage) return false;
    return getChannel(b).then(function (c) {
      var d = b.state.garbage;
      return enqueueInbox(a, d.content, d.properties).then(function (a) {
        debugx.enabled && debugx('recoverGarbage() - ack()');
        c.ack(b.state.garbage);
        b.state.garbage = undefined;
        return true;
      });
    });
  });
};

var requeueGarbage = garbageAction['requeue'] = function (a) {
  a = a || this;
  return resolveTrash(a).then(function (b) {
    if (!b.state.garbage) return false;
    return getChannel(b).then(function (c) {
      var d = b.state.garbage;
      return enqueueTrash(a, d.content, d.properties).then(function (a) {
        debugx.enabled && debugx('requeueGarbage() - ack()');
        c.ack(b.state.garbage);
        b.state.garbage = undefined;
        return true;
      });
    });
  });
};

Handler.prototype.examine = function (a) {
  var b = this;
  var c = { obtained: 0 };
  c.callback = zapper.isFunction(a);
  return examineGarbage(b).then(function (d) {
    if (d) {
      c.obtained = 1;
      if (!c.callback) return c;
      return new Promise(function (e, f) {
        var g = lodash.pick(d, ['content', 'fields', 'properties']);
        a(g, function a(g, h) {
          if (zapper.isObject(h)) {
            d.content = zapper.bufferify(h.content);
            if (zapper.isObject(h.properties)) {
              lodash.merge(d.properties, h.properties);
            }
          }
          if (zapper.isFunction(garbageAction[g])) {
            c.nextAction = true;
            garbageAction[g](b).then(function () {
              e(c);
            }).catch(function (a) {
              c.actionError = a;
              f(c);
            });
          } else {
            c.nextAction = false;
            e(c);
          }
        });
      });
    } else {
      return c;
    }
  });
};

Handler.prototype.close = function () {
  var a = this;
  return Promise.resolve().then(function () {
    return stopPublisher({ state: getProducerState(a) });
  }).then(function () {
    return stopPublisher({ state: getPipelineState(a) });
  }).then(function () {
    return stopSubscriber({ state: getInboxState(a) });
  }).then(function () {
    return stopSubscriber({ state: getTrashState(a) });
  }).then(function () {
    return stopSubscriber({ state: getChainState(a) });
  });
};

Handler.prototype.destroy = Handler.prototype.close;

var b = function b(a) {
  a = a || this;
  return Promise.resolve().then(function () {
    if (a.state.channel) {
      var b = Promise.promisify(a.state.channel.close, {
        context: a.state.channel
      });
      return b().then(function (b) {
        delete a.state.channel;
        debugx.enabled && debugx('destroy() - channel is closed: %s', JSON.stringify(b));
        return true;
      });
    }
    return true;
  }).then(function () {
    if (a.state.connection) {
      var b = Promise.promisify(a.state.connection.close, {
        context: a.state.connection
      });
      return b().then(function (b) {
        delete a.state.connection;
        debugx.enabled && debugx('destroy() - connection is closed: %s', JSON.stringify(b));
        return true;
      });
    }
    return true;
  });
};

var b = function b(a) {
  a = a || this;
  return Promise.resolve().then(function () {
    if (a.state.channel && a.state.consumerRefs && a.state.consumerRefs.length > 0) {
      debugx.enabled && debugx('stopSubscriber() - consumers cancellation has been invoked');
      var b = Promise.promisify(a.state.channel.cancel, {
        context: a.state.channel
      });
      return Promise.mapSeries(a.state.consumerRefs, function (a) {
        return b(a.consumerTag).then(function (a) {
          debugx.enabled && debugx('stopSubscriber() - consumer is cancelled: %s', JSON.stringify(a));
          return a;
        });
      });
    }
    return true;
  }).then(function () {
    if (a.state.channel) {
      var b = Promise.promisify(a.state.channel.close, {
        context: a.state.channel
      });
      return b().then(function (b) {
        delete a.state.channel;
        debugx.enabled && debugx('stopSubscriber() - channel is closed: %s', JSON.stringify(b));
        return true;
      });
    }
    return true;
  }).then(function () {
    if (a.state.connection) {
      var b = Promise.promisify(a.state.connection.close, {
        context: a.state.connection
      });
      return b().then(function (b) {
        delete a.state.connection;
        debugx.enabled && debugx('stopSubscriber() - connection is closed: %s', JSON.stringify(b));
        return true;
      });
    }
    return true;
  });
};

module.exports = Handler;